"""
Module to get a random LeetCode question according to certain
parameters and then output a local Python file to work on
said question.
"""

import random
import re
import subprocess
import textwrap
from pathlib import Path
from typing import Literal

import requests
from bs4 import BeautifulSoup

from localeet.language_maps import (
    LANGUAGE_TO_COMMENT,
    LANGUAGE_TO_EXTENSION,
)


ROOT = 'https://leetcode.com'
API_URL = f'{ROOT}/api/problems/all/'
GQL_URL = f'{ROOT}/graphql'

SLUG_KEY = 'question__title_slug'


def query_all_questions() -> dict:
    """Query LeetCode API for index of all questions"""
    return requests.get(API_URL, timeout=5).json()['stat_status_pairs']


def choose_a_valid_question(
        questions: list[dict],
        max_difficulty: Literal[1, 2, 3],
        min_difficulty: Literal[1, 2, 3],
    ) -> dict:
    """Recurse until a valid question is found, return its slug"""
    choice = random.choice(questions)
    if any((
        choice['paid_only'],
        choice['difficulty']['level'] < min_difficulty,
        choice['difficulty']['level'] > max_difficulty,
        choice['stat'].get(SLUG_KEY) is None,
    )):
        return choose_a_valid_question(
            questions,
            max_difficulty,
            min_difficulty,
        )
    return choice['stat'][SLUG_KEY]


def get_question_data(question_slug: dict) -> dict:
    """Get all metadata available for question via GraphQL query"""
    return requests.post(GQL_URL, timeout=10, json={
        'operationName': 'questionData',
        'variables': {
            'titleSlug': question_slug,
        },
        'query': """query questionData($titleSlug: String!) {
                        question(titleSlug: $titleSlug) {
                            questionId
                            questionFrontendId
                            boundTopicId
                            title
                            titleSlug
                            content
                            translatedTitle
                            translatedContent
                            isPaidOnly
                            difficulty
                            likes
                            dislikes
                            isLiked
                            similarQuestions
                            contributors {
                                username
                                profileUrl
                                avatarUrl
                                __typename
                            }
                            langToValidPlayground
                            topicTags {
                                name
                                slug
                                translatedName
                                __typename
                            }
                            companyTagStats
                            codeSnippets {
                                lang
                                langSlug
                                code
                                __typename
                            }
                            stats
                            hints
                            solution {
                                id
                                canSeeDetail
                                __typename
                            }
                            status
                            sampleTestCase
                            metaData
                            judgerAvailable
                            judgeType
                            mysqlSchemas
                            enableRunCode
                            enableTestMode
                            envInfo
                            libraryUrl
                            __typename
                        }
                    }
                """,
    }).json()


def parse_question_details(question_data: dict) -> dict[str, str]:
    """Parse response from GraphQL down into data needed for output"""
    soup = BeautifulSoup(question_data['data']['question']['content'], 'lxml')
    for code_tag in soup.find_all('code'):
        code_tag.replace_with('`' + code_tag.text + '`')
    return {
        'code_snippets': question_data['data']['question']['codeSnippets'],
        'difficulty': question_data['data']['question']['difficulty'],
        'question_id': question_data['data']['question']['questionId'],
        'question': soup.get_text().replace('\u00A0', ' '),
        'test_case': question_data['data']['question']['sampleTestCase'],
        'title': question_data['data']['question']['title'],
    }


def output_code_file(
        output_path: Path,
        question_details: dict[str, str],
        language: str,
    ) -> str:
    """Take question details and output a python file shell"""
    difficulty, qid, question, snippets, test_case, title = (
        question_details['difficulty'],
        question_details['question_id'],
        question_details['question'],
        question_details['code_snippets'],
        question_details['test_case'],
        question_details['title'],
    )
    extension = LANGUAGE_TO_EXTENSION[language]
    oc = LANGUAGE_TO_COMMENT[language]['open_block']
    cc = LANGUAGE_TO_COMMENT[language]['close_block']
    lc = LANGUAGE_TO_COMMENT[language]['line']
    snippet = next(i for i in snippets if i['langSlug'] == language)['code']
    output_path.mkdir(parents=True, exist_ok=True)
    regex = r'[-\s]+'  # replace spaces and hyphens with underscores
    file_name = f'{re.sub(regex, "_", title.lower())}.{extension}'
    output_path = output_path / file_name
    header = f'{oc}\n{qid} - {difficulty} - {title}\n\n{question}\n{cc}\n\n'
    content = header + snippet + f'\n{lc} Example test case:\n'
    content += '\n'.join([f'{lc} {d}' for d in test_case.split('\n')])
    content += f'\n\n{lc} Suggestion to get started:'
    content += f'\n{lc} 1. write 5-10+ good test cases, including edge cases'
    content += f'\n{lc} 2. write code to take test cases and call the function'
    content = [c.rstrip() for c in content.split('\n')]
    wrapped_content = []
    for s in content:
        if not s.strip():
            wrapped_content.append('')
            continue
        wrapped_lines = textwrap.wrap(s, width=79)
        wrapped_content.extend(wrapped_lines)
    content = '\n'.join(wrapped_content) + '\n'
    with output_path.open('w') as f:
        f.write(content)
    return str(output_path)


def open_code_editor(command: str, file_path: str) -> None:
    subprocess.run([command, file_path])  # noqa: S603


def run(
        max_difficulty: Literal[1, 2, 3],
        min_difficulty: Literal[1, 2, 3],
        output_path: Path,
        code_editor_open_command: str,
        programming_language: str,
    ) -> None:
    all_questions = query_all_questions()
    question_slug = choose_a_valid_question(
        all_questions,
        max_difficulty,
        min_difficulty,
    )
    result = get_question_data(question_slug)
    question_details = parse_question_details(result)
    output_path = output_code_file(
        output_path,
        question_details,
        programming_language,
    )
    open_code_editor(code_editor_open_command, output_path)



import json
from pathlib import Path
from typing import Any

import pytest


class MockValue:
    """Allows any of a datatype to pass comparison within dictionary"""
    def __init__(
            self,
            name: str = 'MOCK_VALUE',
            data_type: type = None,
            validator: callable = None,
        ) -> None:
        self.name = name
        self.data_type = data_type
        self.validator = validator

    def __eq__(self, item: Any) -> bool:  # noqa: ANN401
        if (
            (self.data_type and not isinstance(item, self.data_type)) or
            (self.validator and not self.validator(item))
        ):
            return False
        return True

    def __ne__(self, item: Any) -> bool:  # noqa: ANN401
        return not self.__eq__(item)

    def __repr__(self) -> str:
        return f'<ANY_{self.name}>'


@pytest.fixture(scope='session')
def any_int() -> MockValue:
    return MockValue('INT', int)


@pytest.fixture(scope='session')
def any_str() -> MockValue:
    return MockValue('STR', str)


@pytest.fixture(scope='session')
def any_version() -> MockValue:
    parts_in_semantic_version = 3
    return MockValue(
        'VERSION_NUMBER',
        str,
        lambda x: len(x.split('.')) == parts_in_semantic_version,
    )


@pytest.fixture(scope='session')
def any_json_str() -> MockValue:
    return MockValue('JSON_STR', str, lambda x: json.loads(x))


@pytest.fixture(scope='session')
def two_sum_details_json() -> dict:
    with Path('tests/data/two_sum_details.json').open() as f:
        return json.load(f)


@pytest.fixture(scope='session')
def two_sum_essentials() -> dict:
    with Path('tests/data/two_sum_essentials.json').open() as f:
        return json.load(f)



"""
Tests that version number has increased from PyPI-deployed version.
"""

from localeet.get_version import get_version

from conftest import MockValue


def test_get_version(any_version: MockValue) -> None:
    assert get_version() == any_version



import json
from pathlib import Path

import pytest

from localeet.get_leetcode_problem import (
    choose_a_valid_question,
    get_question_data,
    output_code_file,
    parse_question_details,
    query_all_questions,
)
from localeet.language_maps import LANGUAGE_TO_EXTENSION


NUMBER_OF_QUESTIONS_WHEN_WRITING = 2787


def test_query_all_questions(any_int):
    """Test that query for question list works"""
    result = query_all_questions()
    assert len(result) >= NUMBER_OF_QUESTIONS_WHEN_WRITING
    assert result[-1] == {  # assuming order never changes...
        'difficulty': {'level': 1},
        'frequency': 0,
        'is_favor': False,
        'paid_only': False,
        'progress': 0,
        'stat': {
            'frontend_question_id': 1,
            'is_new_question': False,
            'question__article__has_video_solution': True,
            'question__article__live': True,
            'question__article__slug': 'two-sum',
            'question__hide': False,
            'question__title': 'Two Sum',
            'question__title_slug': 'two-sum',
            'question_id': 1,
            'total_acs': any_int,
            'total_submitted': any_int,
        },
        'status': None,
    }


def test_choose_a_valid_question():
    result = choose_a_valid_question(query_all_questions(), 1, 1)
    assert isinstance(result, str)


def test_get_question_data(
        two_sum_details_json,
        any_int,
        any_json_str,
        any_str,
    ):
    two_sum_details_json['data']['question'].update({
        'dislikes': any_int,
        'judgeType': any_str,
        'likes': any_int,
        'stats': any_json_str,
    })
    test_results = get_question_data('two-sum')
    try:
        assert test_results == two_sum_details_json
    except AssertionError:
        print('Saving details.json to compare differences')
        with Path('details.json').open('w') as f:
            json.dump(test_results, f, indent=4)
        raise


def test_parse_question_details(two_sum_details_json, two_sum_essentials):
    test_results = parse_question_details(two_sum_details_json)
    try:
        assert test_results  == two_sum_essentials
    except AssertionError:
        print('Saving essentials.json to compare differences')
        with Path('essentials.json').open('w') as f:
            json.dump(test_results, f, indent=4)
        raise


@pytest.mark.parametrize('language', ['python', 'rust', 'golang'])
def test_output_python_file(
        two_sum_essentials,
        language,
    ):
    extension = LANGUAGE_TO_EXTENSION[language]
    file_name = f'two_sum.{extension}'
    with Path(f'tests/data/{file_name}').open() as f:
        expected = f.read()
    path = Path('.')
    new_file = path / file_name
    output_path = output_code_file(path, two_sum_essentials, language)
    with new_file.open() as f:
        new_file_contents = f.read()
    try:
        assert output_path == file_name
        assert new_file_contents == expected
    except AssertionError:
        print(f'Saving {file_name} to compare differences')
        raise
    else:
        new_file.unlink()



"""
Entry point for CLI. Define and parse CLI arguments.
"""
from os import environ as env
from pathlib import Path
from typing import Optional

import click

from localeet.get_leetcode_problem import run
from localeet.get_version import get_version
from localeet.language_maps import LANGUAGE_TO_EXTENSION


DIFFICULTY_MAP = {
    'easy': 1,
    'medium': 2,
    'hard': 3,
    '1': 1,
    '2': 2,
    '3': 3,
}

SUPPORTED_LANGUAGES = list(LANGUAGE_TO_EXTENSION.keys())


@click.command()
@click.option(
    '--max_difficulty', '--max',
    help='Max difficulty allowed',
    type=click.Choice(list(DIFFICULTY_MAP.keys())),
    default=env.get('LOCALEET_DEFAULT_MAX_DIFFICULTY', 'hard'),
)
@click.option(
    '--min_difficulty', '--min',
    help='Min difficulty allowed',
    type=click.Choice(list(DIFFICULTY_MAP.keys())),
    default=env.get('LOCALEET_DEFAULT_MIN_DIFFICULTY', 'easy'),
)
@click.option(
    '--output_path', '--path', '-o',
    help='Output path for code file. Will create new directories as needed.',
    default=env.get('LOCALEET_DEFAULT_OUTPUT_PATH', '.'),
)
@click.option(
    '--code_editor_open_command', '--editor', '-e',
    help='Will open the specified editor on the created file.',
    default=env.get('LOCALEET_DEFAULT_CODE_EDITOR_OPEN_COMMAND', 'code'),
)
@click.option(
    '--programming_language', '--language', '-l',
    help='The programming language you want to use for your output file',
    type=click.Choice(SUPPORTED_LANGUAGES),
    default=env.get('LOCALEET_DEFAULT_LANGUAGE', 'python3'),
)
@click.option(
    '--version', '-v',
    help='Print the current version of localeet',
    is_flag=True,
)
def main(
        max_difficulty: str,
        min_difficulty: str,
        output_path: str,
        code_editor_open_command: str,
        programming_language: str,
        version: bool,
    ) -> Optional[str]:
    """Entry point for CLI. Parse CLI arguments."""
    if version:
        return get_version()

    max_difficulty = DIFFICULTY_MAP.get(max_difficulty)
    min_difficulty = DIFFICULTY_MAP.get(min_difficulty)

    output_path = Path(output_path)

    language = programming_language.lower()
    if SUPPORTED_LANGUAGES.index(language) is None:
        msg = f'{programming_language} is not a supported languge'
        raise ValueError(msg)
    if language == 'python':
        language = 'python3'  # python2 is dead, long live python3
    elif language == 'go':
        language = 'golang'

    run(
        max_difficulty,
        min_difficulty,
        output_path,
        code_editor_open_command,
        language,
    )
    return None
